/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@francetelecom.com)
 * Date: 14 sept. 2004
 * Time: 11:15:44
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbJoin;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbTable;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

public class RdbViewGenInfos {
    public Logger logger;
    public boolean debug = false;

    RdbViewGenInfos(Logger logger) {
        this.logger = logger;
        debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
    }

    /**
     * Get the name of the view to use when creating the extent view for a given
     * project and for a given mapper
     * @param clazz the class fro which getting the view name
     * @param projectName the name of the project
     * @param mapperName the name of the mapper
     * @return the name of the view
     */
    public String getViewName(Class clazz, String projectName,
                                               String mapperName) {
        RdbClassMapping cm = 
                (RdbClassMapping) clazz.getClassMapping(projectName, mapperName);
        return cm.getViewName();
    }

    /**
     * Calculates the view representing the extent of the class.
     * The view will be made of an union (in case of horizontal mappings)
     * of filtered (in case of filtered and vertical mappings) tables.
     * Elements of the extent may come from the tables on which the class is mapped
     * or from tables on which its sub(sub...)classes are mapped.
     * The main problem is to retrive those tables and in those tables the right
     * columns mapping fields defined in the mother class knowing that:
     *   - different inheritance rule may me encountered (filtered, horizontal,
     *     vertical). Some will need union, some others will need selection
     *   - classes my be mapped on multiple tables, requiring joins to retrieve the
     *     needed columns
     *   - only tables columns mapping fields of the mother class should be include
     *     in the expression of the view
     *   - columns mapping the same field in different class mapping may have
     *     different names, thus requiring renaming for realizing the required unions
     *
     * How the view is calculated depends on whether the class is at the top
     * of the hierarchy tree.
     * 1) If is at the top, its extent is made of all the instances comming from its
     *    tables and those of its children (if they are mapped to other tables).
     * 2) If it is in the middle of the tree and its inheritance rule is 'to-new-structures'
     *    then, none of its instances comes from its parents' tables, so we can apply the
     *    same algorithm
     * 3) If it is in the middle and its inheritance rule is 'to-extended-structures', then its
     *    filter and those of its children must be used not to get instances of its
     *    parent class.
     *
     * In the first two cases, the view is created with the following formula:
     *   V(A) = T(A) U (U (PROJECT(ATTR(A)) T))
     *               |
     *      T in T(SUB(A)) and T != T(A)
     * where V(A) = view for A's extent, ATTR(A) = attributes of A,
     *       T(X) = main table for class X, SUB(A) = all sub(sub(...))classes of A
     *
     * In the second case, the view is created with the following formula:
     *   V(A) = (FILTER(A) AND (AND FILTER(X))) T(A) U (U (PROJECT(ATTR(A)) T(X)))
     *                           |                      |
     *     X in SUB_F(A) and T(X)=T(A)   T in T(SUB(A) - SUB_F(A)) and T != T(A)
     * where V(A) = view for A's extent, ATTR(A) = attributes of A,
     *       T(X) = main table for class X,
     *       SUB_F(A) = all sub(sub(...))classes that is 'filter chained'* w/ A
     *    (*) 'filter chained' means that all inheritance link from A to X
     *         is a filter inheritance rule
     *
     * Actually, this is a simplification since a single class my be mapped to
     * multiple tables and the extent be calculated by joining these tables.
     * However, since the filters express conditions on the key and the key is
     * entirely contained in a table, this does not change the principles
     * discussed above but simply complexifies the expression of the formulae.
     * Instead of T(A) (resp. T(X)) we will have the joins between A's
     * (resp. X's) tables.
     *
     * Also, the case of abstract classes must be taken into account. If class A
     * its extents is the union of the extents of its subclasses. This is true
     * whether it defines or extends an inherited mapping or not. So there is
     * no need to construct a "select from where" at its level.
     *
     * @return a collection of ExtentMappingInfos that must be unioned to get the extent
     */

    public Collection getPolymorphicClassExtent(Class classA, String projectName,
                                               String mapperName) throws PException {
        //System.out.println("debug=" + debug + "logger=" + logger
        //    + (logger == null ? "" : ", logger.name=" + logger.getName()));
        if (debug) logger.log(BasicLevel.INFO, "Calculating extent view for class"
                    + classA.getFQName() + " project: " + projectName +
                ", mapperName: " + mapperName);
        RdbClassMapping classMappingOfA = (RdbClassMapping)
                classA.getClassMapping(projectName, mapperName);


        // The term "class A" is used for the name of the class of which we are calculating
        // the extent view and "class X" is used as a sub(sub...)class of A

        // get all the primitive elements of class A (non-hidden)
        ArrayList primElsOfA =  new ArrayList();
        if (debug) logger.log(BasicLevel.DEBUG, "Primitive elements are: ");
        for (Iterator itTE = classA.getAllFields().iterator(); itTE.hasNext();) {
            Object te = itTE.next();
            if (te instanceof PrimitiveElement) {
                primElsOfA.add(te);
                if (debug) logger.log(BasicLevel.DEBUG, "   " +
                        ((PrimitiveElement) te).getName());
            }
        }
        // add hidden fields
        for (Iterator itTE = classA.getAllHiddenFields().iterator(); itTE.hasNext();) {
            Object te = itTE.next();
            if (te instanceof PrimitiveElement) {
                primElsOfA.add(te);
                if (debug) logger.log(BasicLevel.DEBUG, "   " +
                        ((PrimitiveElement) te).getName() + " (hidden)");
            }
        }

        HashMap extMappingInfosMap = new HashMap(); // <Class->ExtentMappingInfos>
        HashSet extMappingInfos = new HashSet(); // <ExtentMappingInfos>

        // Remaining classes to inspect. The algorithm follows a top-down
        // horizontal traversal of the tree, classes being added progressively
        // during the traversal
        ArrayList remClasses = new ArrayList();
        remClasses.add(classA);
        while (!remClasses.isEmpty()) {
            // remove the first class in list to treat it
            Class classX = (Class) remClasses.remove(0);
            // and add its children
            Collection xSubClasses = classX.getSubClasses();
            if (xSubClasses != null) {
                remClasses.addAll(xSubClasses);
            }

            // get its class mapping in the current project
            RdbClassMultiMapping classMappingOfX = (RdbClassMultiMapping)
                    classX.getClassMapping(projectName, mapperName);
            if (classMappingOfX == null) {
                throw new PException("Could not find a class mapping for class "
                        + classA.getName());
            }


            // will reference the extent mapping information relative to class X
            // EMIs can be shared by classes that map fields to the same structures
            ExtentMappingInfos emi;


            // sub-class inspection
            if (classX != classA) {
                if (debug) logger.log(BasicLevel.DEBUG, "Inspecting subclass: "
                            + classX.getName());
                Class[] parent = new Class[1];
                ExtentMappingInfos emiParent = getParentEMI(classX, extMappingInfosMap, parent);
                // NB: the ext mapping info may be attached to a ancestor of the parent
                // so do not use emi.clazz to retrieve the pcm but the real parent instead
                ParentClassMapping pcmOfX = classMappingOfX.
                        getParentClassMapping(parent[0].getFQName());

                // sanity check
                if (pcmOfX == null) {
                    throw new PException("Could not find a parent class mapping" +
                            " for class " + classX.getName() + " to super class" +
                            classA.getName());
                }

                // gets the inheritance rule
                String ruleName = pcmOfX.getRuleName();
                if (debug) logger.log(BasicLevel.DEBUG, "  inherits mapping from : " +
                        ((Class) pcmOfX.getLinkedMO()).getFQName() +
                        " with rule=" + ruleName);


                // in the case of an 'extended' or 'added' mapping,
                // add the filter if it exists (and if required) to the
                // emi of the parent
                if ((RdbClassMapping.MAP_NEW_FIELDS_TO_EXTENDED_STRUCTURES
                        .equalsIgnoreCase(ruleName)) ||
                        (RdbClassMapping.MAP_NEW_FIELDS_TO_ADDED_STRUCTURES
                        .equalsIgnoreCase(ruleName))) {
                    // if the parent has unmapped fields, then we need to
                    // construct the union of its children's extents
                    // Otherwise, we can generate a single select/from/where
                    // clause with approriate filters
                    if (emiParent.hasUnmappedFields) {
                        if (debug) logger.log(BasicLevel.DEBUG,
                                "Parent class has unmapped fields, create new EMI");
                        emi = new ExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                        extMappingInfos.add(emi);
                        emi.addFilters = emiParent.addFilters;
                        emi.hasUnmappedFields = classMappingOfX.
                                hasUnmappedPrimitiveElements(primElsOfA);
                    } else {
                        if (debug) logger.log(BasicLevel.DEBUG,
                                "Parent class has all its fields mapped, complement its EMI");
                        emi = emiParent;
                    }

                    // add filter if required and if some exist (there is none
                    // if the class is abstract)
                    if (emi.addFilters && (!classX.isAbstract())) {
                        emi.filters.add(classX.getFQName());
                    }
                } else if (RdbClassMapping.REMAP_FIELDS_TO_NEW_STRUCTURES
                        .equalsIgnoreCase(ruleName)) {
                    // mapping completely redefined on new structures,
                    // so create a new extent mapping info to generate a seperate
                    // select/from/where clause
                    emi = new ExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                    // TODO: should optimize this. The necessity to add filters depends on the inheritance
                    // relationship with the super classes AND the sub-classes
                    emi.addFilters = true;
                    extMappingInfos.add(emi);
                    emi.hasUnmappedFields = classMappingOfX.hasUnmappedPrimitiveElements(primElsOfA);
                } else {
                    throw new InternalError("Unexpected inheritance rule: "
                            + ruleName);
                }
                // register the extent mapping info for that class
                extMappingInfosMap.put(classX, emi);

                // if class A is abstract it has no instances and has
                // (or at least should have) no filters
                // so we can stop here
                if (classX.isAbstract()) {
                    if (debug) logger.log(BasicLevel.DEBUG, "Class is abstract, next one");
                    continue;
                }

            } else { // class X == class A
                emi = new ExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                extMappingInfos.add(emi);
                extMappingInfosMap.put(classX, emi);

                // in case of class A, determines the 'add filters' condition
                // that will be checked for its subclasses
                // TODO: should optimize this. The necessity to add filters depends on the inheritance
                // relationship with the super classes AND the sub-classes
//                emi.addFilters = classMappingOfX.inheritsStructures();
                emi.addFilters = true;
                emi.hasUnmappedFields = classMappingOfX.hasUnmappedPrimitiveElements(primElsOfA);
                if (debug) logger.log(BasicLevel.DEBUG, "  should add filters ? "
                        + (emi.addFilters ? "yes" : "no"));

                // if class A is abstract it has no instances and has
                // (or at least should have) no filters
                // so we can stop here
                if (classX.isAbstract()) {
                    if (debug) logger.log(BasicLevel.DEBUG, "Class is abstract, next one");
                    continue;
                }
                if (emi.addFilters) {
                    emi.filters.add(classX.getFQName());
                }
            }



            // finds how primitive elements defined in class A are mapped
            // in class X and construct select/from/where clauses to fetch
            // the corresponding tuples from X's tables
            int pos = 0;
            for (Iterator it = primElsOfA.iterator(); it.hasNext(); pos++) {
                PrimitiveElement pe = (PrimitiveElement) it.next();
                if (debug) logger.log(BasicLevel.DEBUG, "     find mapping for "
                        + "primitive element " + pe.getName());
                if (emi.mappingDone(pos)) {
                    if (debug) logger.log(BasicLevel.DEBUG, "    already done");
                    continue;
                }
                // search the PEM, looking in super classes if necessary
                RdbPrimitiveElementMapping mappingOfpeInX = (RdbPrimitiveElementMapping)
                        classMappingOfX.getPrimitiveElementMapping(pe.getName(), true);
                // for classes mapped to several tables, the primitive element
                // may be mapped on an external table, requiring a join
                RdbJoin join = mappingOfpeInX.getJoinByPrimitiveElement(pe);
                String tableName;
                if (join != null) {
                    if (debug) logger.log(BasicLevel.DEBUG, "     mapped to external table");
                    tableName = join.getExternalTable().getName();
                    emi.joins.add(join);
                } else {
                    if (debug) logger.log(BasicLevel.DEBUG, "     mapped to main table");
                    emi.mainTable = (RdbTable) mappingOfpeInX.getParent();
                    tableName = emi.mainTable.getName();
                }

                emi.addProjection(tableName, mappingOfpeInX.getName(),
                        pe.getName(), pos);

            }
        } // while !remClasses.isEmpty()

        // remove empty requests
        for (Iterator emiIt = extMappingInfos.iterator(); emiIt.hasNext();) {
            ExtentMappingInfos emi = (ExtentMappingInfos) emiIt.next();
            if (emi.isEmpty()) {
                emiIt.remove();
            }
        }

    return extMappingInfos;
    }
    /**
     * finds the parent's extent mapping info of a class in the inheritance hierarchy
     * @param clazz the clazz from wich searching for parents
     * @param EMIMap map of < clazz,extent mapping infos > into wich searching
     * @return
     */
    private ExtentMappingInfos getParentEMI(Class clazz, Map EMIMap, Class[] parent) {
        logger.log(BasicLevel.DEBUG, "Get parent's EMI");
        if (parent != null) {
            parent[0] = null;
        }
        ExtentMappingInfos emiOfParent = null;
        Collection parentsOfX = clazz.getSuperClasses();
        for (Iterator itP = parentsOfX.iterator(); itP.hasNext();) {
            Class superClass = (Class) itP.next();
            emiOfParent = (ExtentMappingInfos)
                    EMIMap.get(superClass);
            if (emiOfParent != null) {
                if (parent != null) {
                    logger.log(BasicLevel.DEBUG, "Return parent=" + superClass.getFQName());
                    parent[0] = superClass;
                }
                return emiOfParent;
            }
        }
        logger.log(BasicLevel.DEBUG, "Found none");
        return null;
    }
    /**
     * This class is used to construct extent mapping information when calculating
     * views (see getNonFinalClassExtent) and that will then be used to
     * generate the appropriate union of select/from/where clauses
     */
    public class ExtentMappingInfos {

        // class to which the expression is attached
        Class clazz;
        // names of columns to which fields are mapped
        private String[] column_names;
        // alias given to the projections
        private String[] column_aliases;

        // joins required to get the desired columns
        private HashSet joins = new HashSet(); // {RdbJoin}
        // tables required to get the desired columns
        private RdbTable mainTable;

        // the filter set will contain the name of classes from which filtering
        // expressions must be retrived so as to be added to the where clause.
        HashSet filters = new HashSet();  // {String}
        // if addfilters is true, then filters of classes mapped on the same
        // tables (extended/added inheritance) must be added
        boolean addFilters = false;
        boolean hasUnmappedFields = false;
        private boolean emptyEMI = true;
        
        // we can give a name to EMIs
        private String name;
        
        ExtentMappingInfos(Class clazz, String name, int nbColumns) {
            this.clazz = clazz;
            this.name = name;
            this.column_names = new String[nbColumns];
            this.column_aliases = new String[nbColumns];
        }

		/**
		 * @return Returns the name.
		 */
		public String getName() {
			return name;
		}
		/**
		 * @param name The name to set.
		 */
		public void setName(String name) {
			this.name = name;
		}
        // getters required by Velocity
        public HashSet getFilters() {
            return filters;
        }

        public RdbTable getMainTable() {
            return mainTable;
        }

        public HashSet getJoins() {
            return joins;
        }

        public boolean isEmpty() {
            return emptyEMI;
        }

        public void addProjection(String tableName, String columnName,
                                      String aliasName, int position) {
            if (column_names[position] != null) {
                if (debug) {
                    String newExpr = tableName + "." + columnName;
                    if (!column_names[position].equals(newExpr)) {
                        throw new InternalError("expr in select definition (" + newExpr
                                + ") not equal to old one (" + column_names[position] + ")");
                    }
                    if (!column_aliases[position].equals(aliasName)) {
                        throw new InternalError("alias in select definition (" + aliasName
                                + ") not equal to old one (" + column_aliases[position] + ")");
                    }
                }
            } else {
                emptyEMI = false;
                column_names[position] = tableName + "." + columnName;
                column_aliases[position] = aliasName;
            }
        }


        /**
         * get a select parameter to give to the getQuery method of adapter
         * this method is called in RdbViewDefinition.vm when generating the
         * view definition
         * @param adapterName the name of the adapter in the generated code
         * @return a string to include as the first parameter of the call to
         * getQuery in the generated code
         */
        public String getSelectParameter(String adapterName) {
            StringBuffer sb = new StringBuffer();
            // select parameter
            for (int pos = 0; pos < column_names.length; pos++) {
                sb.append("\"" + column_names[pos] + "\" + " +
                        adapterName +".getColumnAliasExpr(\"" + column_aliases[pos] + "\")");
                if (pos < column_names.length - 1) {
                    sb.append(" + \",\" + ");
                }
            }
             return sb.toString();
        }

        public boolean mappingDone(int pos) {
            return (column_names[pos] != null);
        }
        public boolean hasJoins() {
            return (!joins.isEmpty());
        }
    }

}
