/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.lib.AbstractPMapcluster;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.HashSet;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * Defines the PMapCluster used to manage relational schema for the RDB
 * mappers.
 * @author	P. Dechamboux
 */
public class RdbPMapCluster extends AbstractPMapcluster {

    
	/**
	 * The list of table managed by this PMapCluster. It contains Table instance.
	 */
	private ArrayList tables;
    /**
     * Views indexed with their names < String -> View >
     */
	private Map views = new HashMap();
    
	private RdbPMappingStructuresManager manager;
    private RdbAdapter adapter;
	/**
	 * Constructs a RdbPMapCluster for an abstract class.
	 * @param jcname	The first JORM class name to be assigned to this cluster.
	 * @param man		The manager associated to these mapping structures.
	 */
	RdbPMapCluster(String jcname, RdbPMappingStructuresManager man) {
		super();
		manager = man;
		jcNames = new HashSet(1);
		unresolvedDependencies.add(jcname);
		tables = new ArrayList(1);
		structuresActive = false;
        if (this.manager.getPMapper() != null) {
            adapter = ((PMapperRdb) this.manager.getPMapper()).getRdbAdapter();
        }
	}

	/**
	 * Constructs a RdbPMapCluster.
	 * @param jcname	The first JORM class name to be assigned to this cluster.
	 * @param tname		The first table name to be assigned to this cluster.
	 * @param man		The manager associated to these mapping structures.
	 */
	RdbPMapCluster(String jcname, String tname, RdbPMappingStructuresManager man) {
		this(jcname, man);
		tables.add(new Table(tname, this));
	}


	/**
	 * Defines a new table column into this map cluster if it does not already
	 * exist. If it exist, it verifies that the redefinition is compatible with
	 * existing information.
	 * @param tname		The table name.
	 * @param cname		The column name.
	 * @param type		The SQL type of this column.
	 * @param notnull	The notnull status of this column.
	 * @throws PException Thrown if the column is redefined with incompatible
	 * information.
	 */
	void addTableColumn(String tname, String cname, String type,
						boolean notnull, boolean ispkcol, boolean isMaster) throws PException {
		Table curt = null;
		Iterator it = tables.iterator();
		while (it.hasNext()) {
			curt = (Table) it.next();
			if (curt.name.equals(tname)) {
				break;
			}
			curt = null;
		}
		if (curt == null) {
			// no such table defined yet: create a new one
			curt = new Table(tname, this);
			tables.add(curt);
		}
		Column c = curt.getColumn(cname);
		if (c == null) {
			curt.addColumn(cname, type, notnull, ispkcol, isMaster);
		} else if (isMaster) {
			if (c.isMaster) {
				if (c.type != type) {
					if (ispkcol && !c.isPkCol) {
						c.notNull = notnull;
						c.isPkCol = ispkcol;
						c.type = type;
					} else if (!ispkcol && c.isPkCol) {
						//nothing to do
					} else {
						throw new PException(
								"Incompatible column definition: existing: " + c
								+ " asked : Column (name:" + cname
								+ ", type:" + type
								+ ", notNull:" + notnull
								+ ", isMaster:" + isMaster
								+ ", isPkCol:" + ispkcol
								+ ")");
					}
				}
			} else {
				c.isMaster = true;
				c.notNull = notnull;
				c.isPkCol = ispkcol;
				c.type = type;
			}
		}
	}
        
    /**
     * Add a statement to a view. The view will be the union of all added statements. 
     * Create the view if it does not already exist. The statement is of the
     * form select ... from ... [where ...]. Statements are named so one can register filters to them. 
     * The last parameter indicates whether the statement has a where clause so as to know if filters must
     * be precceded by a AND (if there is already a where clause) or a WHERE (if there no where clause yet) 
     * @param viewName the name of the view
     * @param stmtName the name of the statement
     * @param stmtValue tht select/from/where to be added
     * @param hasWhereClause if true the statement has a where clause
     * @param clazz the class of which we are calculating the extent on which 
     * the appendExtentFilter method will be called
     */
    void addViewStatement(String viewName, String stmtName, String stmtValue, 
            boolean hasWhereClause, RdbPPolymorphicClass clazz) {
        getView(viewName).addStatement(stmtName, stmtValue, hasWhereClause, clazz);        
    }



	/**
	 * Looks for this specified column into this map cluster.
	 * @param tname		The table name.
	 * @param cname		The column name.
	 * @param type		The SQL type of this column.
	 * @param notnull	The notnull status of this column.
	 * @throws PException Thrown if the column is redefined with incompatible
	 * information.
	 */
	boolean containColumn(String tname, String cname, String type,
						  boolean notnull) throws PException {
		Table curt = null;
		Iterator it = tables.iterator();
		while (it.hasNext()) {
			curt = (Table) it.next();
			if (curt.name.equals(tname)) {
				break;
			}
			curt = null;
		}
		if (curt == null) {
			// no such table defined yet : create a new one
			return false;
		}
		return curt.containColumn(cname, type, notnull);
	}

	/**
	 * Looks for this table name into this map cluster.
	 * @param tname	The table name.
	 * @return true if this table name is already assigned to this cluster.
	 */
	boolean containTable(String tname) {
		Iterator it = tables.iterator();
		while (it.hasNext()) {
			Table cur = (Table) it.next();
			if (cur.name.equals(tname))
				return true;
		}
		return false;
	}

	/**
	 * Looks for this table name into this map cluster.
	 * @param tname	The table name.
	 * @return true if this table name is already assigned to this cluster.
	 */
	Table getTable(String tname) {
		Iterator it = tables.iterator();
		while (it.hasNext()) {
			Table cur = (Table) it.next();
			if (cur.name.equals(tname))
				return cur;
		}
		return null;
	}

	/**
	 * Merges map cluster passed in parameter with this one. Clusters are merged
	 * either because they share a table name or a JORM class name.
	 * @param cl
	 */
	void merge(RdbPMapCluster cl) throws PException {
		jcNames.addAll(cl.jcNames);
		unresolvedDependencies.addAll(cl.getUnResolvedDependencies());
		unresolvedDependencies.removeAll(jcNames);
		for (int i = 0; i < cl.tables.size(); i++) {
			Table cur = (Table) cl.tables.get(i);
			Table t = getTable(cur.name);
			if (t == null) {
				tables.add(cur);
			} else {
				t.merge(cur);
			}
		}
	}
    
    /**
     * Gets a view by its name, creates it if it does not exist
     * @param viewName the name of the view to retrieve
     * @return the view by its name, creates it if it does not exist
     */
    private View getView(String viewName) {
        View view = (View) views.get(viewName);
        if (view == null) {
            view = new View(viewName, this);
            views.put(viewName, view);            
        }
        return view;
    }


	// IMPLEMENTATION OF METHODS FROM THE PMapCluster INTERFACE


	/**
	 * Creates the mapping structures defined by this map cluster.
	 * @param force	If true, the creation of an existing table leads to an exception.
	 * @throws PException   Thrown if it cannot be performed (especially
	 * if all or some of these mapping structures already exist.
	 */
	public void createMappingStructures(boolean force) throws PException {
		if (structuresActive) {
			throw new PException("Cannot change mapping structures while they are under use");
		}
		if (!isDefined()) {
			throw new PException("Cannot create hosting struture: the " +
								 "following dependencies has not been resolved: "
								 + unresolvedDependencies);
		}

        Object conn = manager.getConnection();
		try {
            
            // creates the tables
			Iterator it = tables.iterator();
			while (it.hasNext()) {
				Table table = (Table) it.next();
				if (table.exist(RdbConnectionWrapper.narrow2SQL(conn), adapter)) {
					if (force) {
						throw new PException("Cannot create table " + table.name + ": already exist");
					}
				} else {
					table.create(RdbConnectionWrapper.narrow2SQL(conn));
				}
			}
    
            // creates the views representing the full-extent of classes
            for (Iterator itView = views.values().iterator(); itView.hasNext();) {
                View view = (View) itView.next();
                if (!view.exist(RdbConnectionWrapper.narrow2SQL(conn), adapter))
                view.create(RdbConnectionWrapper.narrow2SQL(conn), (PMapperRdb) manager.getPMapper());
            }
            
		} finally {
			manager.closeConnection(conn);
		}
	}

	/**
	 * Deletes the data that have been inserted into the mapping structures
	 * defined by this map cluster.
	 * @throws PException   Thrown when the data store cannot perform this
	 * operation.
	 */
	public void deleteData() throws PException {
		if (!isDefined()) {
			throw new PException("Cannot remove data: the " +
								 "following dependencies has not been resolved: "
								 + unresolvedDependencies);
		}
		Object conn = manager.getConnection();
		try {
			Iterator it = tables.iterator();
			while (it.hasNext()) {
				Table table = (Table) it.next();
				if (table.exist(RdbConnectionWrapper.narrow2SQL(conn), adapter)) {
					table.deleteData(RdbConnectionWrapper.narrow2SQL(conn));
				} else {
					manager.logger.log(BasicLevel.WARN,
									   "Impossible to delete the data of the table '"
									   + table.name + "': table does not exist.");
				}
			}
		} finally {
			manager.closeConnection(conn);
		}
	}

	/**
	 * Deletes all or part of the mapping structures defined by this map
	 * cluster when they exist. If they contain some data, they are also
	 * deleted.
	 * @throws PException   Thrown when the data store cannot perform this
	 * operation.
	 */
	public void deleteMappingStructures() throws PException {
		if (structuresActive) {
			throw new PException("Cannot change mapping structures while they are under use");
		}
		if (!isDefined()) {
			throw new PException("Cannot delete hosting struture: the " +
								 "following dependencies has not been resolved: "
								 + unresolvedDependencies);
		}
		Object conn = manager.getConnection();
		try {
            // deletes structures according to the dependency views -> tables
            // deletes the views representing the full-extent of classes
            for (Iterator itView = views.values().iterator(); itView.hasNext();) {
                View view = (View) itView.next();
                if (view.exist(RdbConnectionWrapper.narrow2SQL(conn), adapter)) {
                    view.delete(RdbConnectionWrapper.narrow2SQL(conn));
                }
            }
            // deletes tables
            Iterator it = tables.iterator();
			while (it.hasNext()) {
				Table table = (Table) it.next();
				if (table.exist(RdbConnectionWrapper.narrow2SQL(conn), adapter)) {
					table.delete(RdbConnectionWrapper.narrow2SQL(conn));
				}
			}
		} finally {
			manager.closeConnection(conn);
		}
	}

}
/**
 */
class View {
    
    private class ViewStatement {
        String stmt;
        boolean hasWhereClause;
        RdbPPolymorphicClass clazz;
        ViewStatement(String stmt, boolean hasWhereClause, RdbPPolymorphicClass clazz) {
            this.stmt = stmt;
            this.hasWhereClause = hasWhereClause;
            this.clazz = clazz;
        }
    }
    
    private String name;
    private Map statements = new HashMap();
    private RdbPMapCluster cluster;

    public View(String name, RdbPMapCluster cluster) {
        this.name = name;   
        this.cluster = cluster;
    }
    
    /**
     * Gets the name of the view
     * @return the name of the view
     */
    public String getName() {
        return name;
    }

    /**
     * Add a SQL statement to the view. The view will be the union af all added statements. Note that @hasWhereClause 
     * was used to tell if the statement has a where clause but not used anymore. Instead, the statement is
     * parsed to see if there is a 'where' in it. The rationale is that it is not always trivial to know if the 
     * statement has a where clause. For instance, joins may be expressed with a where clause or with JOIN cluases 
     * depending on the DBMS.
     * @param stmtName the name of the statement
     * @param stmtValue the value of the statement 
     * @param hasWhereClause not used anymore  
     * @param clazz the name of the correponding class 
     */
    public void addStatement(String stmtName, String stmtValue, boolean hasWhereClause, RdbPPolymorphicClass clazz) {
        ViewStatement stmt =  new ViewStatement(stmtValue, hasWhereClause, clazz);
        statements.put(stmtName, stmt);
        stmt.hasWhereClause = stmtValue.toLowerCase().matches(".*where.*");
    }
    
    /**
     * Creates the view in the database. 
     * @param connection
     * @param mapper
     * @throws PException
     */
    void create(Connection connection, PMapperRdb mapper) throws PException {
        Statement stmt = null;
        StringBuffer sb = new StringBuffer("CREATE VIEW " + name + " AS (");

        // add the statements one by one, separated with the UNION keyword
        String unionSep = "";
        for (Iterator it = statements.entrySet().iterator(); it.hasNext();) {
            
            sb.append(unionSep);
            sb.append("(");
            Map.Entry entry = (Map.Entry) it.next();
            String stmtName = (String) entry.getKey();
            ViewStatement vs = (ViewStatement) entry.getValue();
            sb.append(vs.stmt);
            if (vs.clazz != null) {
                if (vs.hasWhereClause) {  
                    sb.append(" AND (");
                    vs.clazz.appendExtentFilters(stmtName, sb);
                    sb.append(")");
                } else {
                    sb.append(" WHERE ");
                    vs.clazz.appendExtentFilters(stmtName, sb);
                }
            }
            sb.append(")");
            unionSep = " UNION ";
        }

        sb.append(")");
        try {
            stmt = connection.createStatement();
            stmt.execute(sb.toString());
        } catch (SQLException e) {
            throw new PExceptionIO(e, "Impossible to create the table '"
                    + name + "' of the cluster " + cluster.getClusterClasses()
                    + " : " + sb.toString());
       }
    }
    
    void delete(Connection connection) throws PException {
        Statement stmt = null;
        PException resex = null;
        try {
            stmt = connection.createStatement();
            stmt.execute("DROP VIEW " + name);
        } catch (java.sql.SQLException se) {
            resex = new PExceptionIO(se);
        } finally {
            if (stmt != null) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    if (resex == null) {
                        resex = new PException(e, "Problem while closing statment.");
                    }
                }
            }
            if (resex != null) {
                throw resex;
            }
        }
    }

    
    /**
     * Checks for the existence of a view in the database.
     * @param connection the connection with the datastore to use
     * @param adapter the adapter to the dbms
     * @return true if the view already exists
     * @throws PException
     */
    boolean exist(Connection connection, RdbAdapter adapter) throws PException {
        try {
            return adapter.existView(connection, name);
        } catch (SQLException e) {
            throw new PException(e, "Pb while testing the existence of view " + name);
        }
    }

}


/**
 * Describes a table into a map cluster.
 */
class Table {
	String name;
	ArrayList columns;
	RdbPMapCluster cluster;


	Table(String tname, RdbPMapCluster cluster) {
		name = tname;
		this.cluster = cluster;
		columns = new ArrayList();
	}

	void addColumn(String cname, String type, boolean notnull, boolean ispkcol, boolean ismaster)
			throws PException {
		Column cur;
		cur = new Column();
		cur.name = cname;
		cur.type = type;
		cur.notNull = notnull;
		cur.isPkCol = ispkcol;
		cur.isMaster = ismaster;
		columns.add(cur);
	}

	void merge(Table t) throws PException {
		for (int i = 0; i < t.columns.size(); i++) {
			Column cur = (Column) t.columns.get(i);
			Column c = getColumn(cur.name);
			if (c == null) {
				columns.add(cur);
			} else {
				c.merge(cur);
			}
		}
	}

	boolean containColumn(String cname, String type,
						  boolean notnull) throws PException {
		Column cur;
		Iterator it = columns.iterator();
		while (it.hasNext()) {
			cur = (Column) it.next();
			if (cur.name.equals(cname)) {
				return true;
			}
		}
		return false;
	}

	Column getColumn(String cname) {
		Column cur;
		Iterator it = columns.iterator();
		while (it.hasNext()) {
			cur = (Column) it.next();
			if (cur.name.equals(cname)) {
				return cur;
			}
		}
		return null;
	}

	void create(Connection connection) throws PException {
		Statement stmt = null;
		StringBuffer sb = new StringBuffer("CREATE TABLE " + name + " (");
		try {
			stmt = connection.createStatement();
			String sep = "";
			Iterator it = columns.iterator();
			while (it.hasNext()) {
				sb.append(sep);
				sep = ", ";
				Column col = (Column) it.next();
				sb.append(col.name);
				sb.append(" ");
				sb.append(col.type);
				if (col.notNull && col.type.indexOf("NOT NULL") == -1) {
					sb.append(" NOT NULL");
				}
			}
			sep = "";
			it = columns.iterator();
			while (it.hasNext()) {
				Column col = (Column) it.next();
				if (!col.isPkCol) {
					continue;
				}
				if (sep.equals("")) {
					sb.append(", PRIMARY KEY (");
				}
				sb.append(sep);
				sep = ", ";
				sb.append(col.name);
			}
			if (!sep.equals("")) {
				sb.append(")");
			}
			sb.append(")");
			stmt.execute(sb.toString());
		} catch (SQLException se) {
			throw new PExceptionIO(se, "Impossible to create the table '"
					+ name + "' of the cluster " + cluster.getClusterClasses()
					+ " : " + sb.toString());
		} finally {
			if (stmt != null) {
				try {
					stmt.close();
				} catch (SQLException e) {
				}
			}
		}
	}

	void deleteData(Connection connection) throws PException {
		Statement stmt = null;
		PException resex = null;
		try {
			stmt = connection.createStatement();
			stmt.execute("DELETE FROM " + name);
		} catch (SQLException se) {
			resex = new PExceptionIO(se, "SQL problem while deleting data: class <" + name + ">");
		} finally {
			if (stmt != null) {
				try {
					stmt.close();
				} catch (SQLException e) {
					if (resex == null) {
						resex = new PException(e, "Problem while closing statment.");
					}
				}
			}
			if (resex != null) {
				throw resex;
			}
		}
	}

	void delete(Connection connection) throws PException {
		Statement stmt = null;
		PException resex = null;
		try {
			stmt = connection.createStatement();
			stmt.execute("DROP TABLE " + name);
		} catch (java.sql.SQLException se) {
			resex = new PExceptionIO(se);
		} finally {
			if (stmt != null) {
				try {
					stmt.close();
				} catch (SQLException e) {
					if (resex == null) {
						resex = new PException(e, "Problem while closing statment.");
					}
				}
			}
			if (resex != null) {
				throw resex;
			}
		}
	}

	boolean exist(Connection connection, RdbAdapter adapter) throws PException {
		try {
			return adapter.existTable(connection, name);
		} catch (SQLException e) {
			throw new PException(e, "Pb while testing the existence of table " + name);
		}
	}
}

/**
 * Describes a column into a table description into a map cluster.
 */
class Column {
	String name;
	String type;
	boolean notNull;
	boolean isPkCol;
	boolean isMaster;

	void merge(Column c) throws PException {
		if (type != c.type) {
			throw new PException("Impossible to merge the column " + name
								 + ": column type does not match");
		}
		if (c.isMaster) {
			notNull = c.notNull;
			isPkCol |= c.isPkCol;
		}
	}

	public String toString() {
		return "Column (name:" + name
				+ ", type:" + type
				+ ", notNull:" + notNull
				+ ", isMaster:" + isMaster
				+ ", isPkCol:" + isPkCol
				+ ")";
	}
}
